# encoding: utf8
"""
This module encapsulates useful classes and data for calculating with SI quantities.
Loosely based on the English translation of the SI brochure
<http://www.bipm.org/utils/common/pdf/si_brochure_8_en.pdf>

>>> from si.units.common import *
>>> print (1*bar)*(1*l)
100.0 J
"""
from __future__ import division
import sys

class MakesNoSense(Exception):
	"""Exception raised when impossible operations are attempted on SI objects, like adding seconds to candela."""

class Exponents(dict):
	"""Dictionary mapping symbolic bases to exponents."""
	def __str__(self):
		"""Standard string representation.
		
		>>> print Exponents({'m':1,'s':-2})
		m/s^2
		>>> print Exponents({'s':-1})
		/s
		>>> print Exponents({'a':1,'b':-2,'c':3,'d':-10})
		a c^3/(b^2 d^10)
		"""
		dpos=[x for x in self.items() if x[1]>0]
		dneg=[(x[0],-x[1]) for x in self.items() if x[1]<0]
		dpos.sort(lambda x,y:cmp(x[1],y[1]))
		dneg.sort(lambda x,y:cmp(x[1],y[1]))
		pospart,negpart=[],[]
		for part,d in [(pospart,dpos),(negpart,dneg)]:
			for s,n in d:
				part.append(str(s)+(n!=1 and "^%d"%n or ""))
		r=" ".join(pospart)
		if negpart:
			r+='/'
			if len(negpart)>1: r+='('
			r+=" ".join(negpart)
			if len(negpart)>1: r+=')'
		return r.strip()

class SI(tuple):
	"""SI quantity, storeed as a 2-tuple of value and a tuple storing the dimension (α to η)."""

	symbols=('m','kg','s','A','K','mol','cd')

	value=property(lambda self:self[0])
	dim=property(lambda self:self[1])
	unit=property(lambda self:SI((1,self.dim)))

	@staticmethod
	def _exponents_mul(A,B):
		return tuple([a+b for a,b in zip(A,B)])
	@staticmethod
	def _exponents_div(A,B):
		return tuple([a-b for a,b in zip(A,B)])

	def __add__(self,other):
		""">>> from si.units.common import *
		>>> print 1*m+2*m
		3 m"""
		if self.dim!=other.dim: raise MakesNoSense("Adding non-compatible quantities.")
		return SI((self.value+other.value,self.dim))
	def __sub__(self,other):
		""">>> from si.units.common import *
		>>> print (m+m)-m
		1 m"""
		if self.dim!=other.dim: raise MakesNoSense("Subtracting non-compatible quantities.")
		return SI((self.value-other.value,self.dim))
	def __mul__(self,other):
		""">>> from si.units.common import *
		>>> print 2*m
		2 m
		>>> print m*2
		2 m
		>>> print 2*m*m
		2 m^2"""
		if not hasattr(other,'dim'):
			return SI((self.value*other,self.dim))
		return SI((self.value*other.value,self._exponents_mul(self.dim,other.dim)))
	__rmul__=__mul__
	def __div__(self,other):
		""">>> from si.units.common import *
		>>> print (m*m)/m
		1.0 m
		>>> print (10*m)/2
		5.0 m
		>>> print (10*m)/(5*m)
		2.0"""
		if not hasattr(other,'dim'):
			return SI((self.value/other,self.dim))
		if other.dim==self.dim: return self.value/other.value
		return SI((self.value/other.value,self._exponents_div(self.dim,other.dim)))
	__truediv__=__div__
	def __rdiv__(self,other):
		""">>> from si.units.common import *
		>>> print 5/(1/s)
		5.0 s"""
		if hasattr(other,'dim'):
			raise NotImplementedError("SI units can handle divisions themselves, no need to do this here")
		return SI((other/self.value,tuple(-x for x in self.dim)))
	__rtruediv__=__rdiv__
	def __pow__(self,exp):
		""">>> from si.units.common import *
		>>> print (5*m)**2
		25 m^2"""
		return SI((self.value**exp,tuple(a*exp for a in self.dim)))

	def __cmp__(self,other):
		if self.dim!=other.dim: raise MakesNoSense("Comparing non-compatible quantities.")
		return cmp(self[0],other[0])
	def __nonzero__(self):
		return bool(self.value)

	def using(self,unit):
		"""Get numeric value in given unit.

		>>> from si.units.exotic import *
		>>> (1*m/s).using(attoparsec/microfortnight) # "how much is 1 m/s in attorparsec/microfortnight?"
		39.200466287804247
		"""
		if self.dim!=unit.dim: raise MakesNoSense("The quantity can not be expressed in that unit.")
		return self.value/unit.value

	def basestring(self):
		"""String representation using only powers of base units."""
		return "%s %s"%(self.value,Exponents(zip(self.symbols,self.dim)))
	def __repr__(self):
		return '<SI %s>'%self.basestring()

	def intelligentstring(self):
		"""Return a string representation using compound SI units. The function will try to

		- minimize the sum of absolute values of exponents: (this is only partially achieved)
		>>> from si.units.common import *
		>>> gravity=10*m/s/s; height=5*m; mass=5*kg
		>>> print gravity*height*mass
		250.0 J

		- avoid units with no positive exponents:
		>>> print 5/s
		5.0 Hz
		"""
		def expsum(u): return sum([abs(x) for x in u])
		units=availableUnits()

		decomposition={}

		lastunits=None
		remaining=self.dim
		while lastunits!=remaining:
			lastunits=remaining
			currentexpsum=expsum(remaining)
			for name,unit in units:
				div=self._exponents_div(remaining,unit.dim)
				if expsum(div)+1<currentexpsum or (expsum(div)+1<=currentexpsum and max([x for x in remaining if x])<0 and not max([x for x in div if x]+[1])<0): # by dividing, another exponent is added (thus +1). substitution with equal number of exponents is only done to avoid negative exponents.
					decomposition[name]=decomposition.get(name,0)+1
					remaining=div
					break
		
		for name,value in zip(self.symbols,remaining):
			assert name not in decomposition
			decomposition[name]=value
		
		return "%s %s"%(self.value,Exponents(decomposition))
				
	__str__=intelligentstring


def availableUnits():
	"""List of name/value tuples for intelligentstring; depends on loaded modules (no units will be used whose module is not yet imported)"""
	intelligentStringUnits=["Ohm", "S","J","W","Pa","N","V","C","Wb","Ohm","H","F","T","Hz"]
	found={}
	for name,module in sys.modules.items():
		if name.startswith('si.units.') and module:
			for u in intelligentStringUnits:
				if u in module.__dict__: found[u]=module.__dict__[u]
	return [(u,found[u]) for u in intelligentStringUnits if u in found]



# encoding: utf8
"""Prefixed units, mostly from base and derived units."""
from si import prefixes
import si.units.base as u_base
import si.units.derived as u_derived
import si.units.nonsi as u_nonsi # only bar

units = {}
units['base'] = u_base.__dict__.items()
units['derived'] = u_derived.__dict__.items()
units['nonsi'] = u_nonsi.__dict__.items()
del u_base, u_derived, u_nonsi
units = dict(units['base']+units['derived']+units['nonsi'])

# prefixes chosen are quite arbitrary, based on what i know to be used, and si recommendations
_3base = list('YZEPTGMkmunpfazy')
prefixes_for_units = {
		'm':_3base+list('cd'),
		's':list('munpfazy'),
		'g':list('YZEPTGMdcmunpfazy')+['da'],
		'A':_3base,
		'K':list('munpfazy'), # others "not recommended"
		'mol':_3base,
		'cd':_3base,
		'l':list('hmunpfazy'),
		'Hz':list('YZEPTGMk'),
		'N':_3base+['da'],
		'Pa':_3base,
		'J':_3base,
		'W':_3base,
		'V':_3base,
		'C':_3base,
		'Wb':_3base,
		'Ohm':_3base,
		'H':_3base,
		'F':_3base,
		'T':_3base,
		'bar':_3base,
		}

_l = locals()
for unit,unit_prefixes in prefixes_for_units.items():
	for prefix in unit_prefixes:
		_l[prefix+unit] = units[unit] * getattr(prefixes,prefix)

for prefix in _3base:
	_l[prefix+'m2'] = _l[prefix+'m']**2
	_l[prefix+'m3'] = _l[prefix+'m']**3

del units, prefixes_for_units, unit, unit_prefixes, prefix, prefixes



# encoding: utf8
"""Units derived from SI base units."""
from si.units.base import *

# as by si definition (table 3)
rad = m/m
sr = m**2/m**2
Hz = s**-1
N = kg*m/s/s
Pa = N/m/m
J = N*m
W = J/s
C = s*A
V = W/A
F = C/V
Ohm = V/A # would have to be Ω or Ω
S = A/V
Wb = V*s
T = Wb/m/m
H = Wb/A
# degree Celsius causes much headache
lm = cd*sr
lx = lm/m**2
Bq = s**-1
Gy = J/kg
Sv = J/kg
kat = mol/s


# accepted non-si units (by si definition) (table 6)
min = 60*s
h = 60*min
d = 24*h
# angles left out; should be °, ′, ″
ha = 10**4*m**2
l = 10**-3*m**3
t = 10**3*kg

# for use with prefixes or general usability
g = kg/1000
m2 = m**2
m3 = m**3



# encoding: utf8
"""Units accepted for SI use"""
from si.units.derived import *
from si import prefixes as _prefixes
import math as _math

# by si description (table 8)
bar = 10**5*Pa
mmHg = 133.322*Pa # roughly
Angrstrom = 10**-10*m # should be Å or Å
M = 1852*m # nautical mile
b = 10**-28*m2 # barn
kn = 1852/3600*m/s # knot
# neper, bel and decibel not implemented lacking understanding of the problems (and/or time)

# CGS system
cm = m/100 # CGS is based on cm, adding it here
erg = 10**-7*J
dyn = 10**-5*N
P = dyn*s/cm/cm
St = cm*cm/s
sb = cd/cm/cm
ph = cd*sr/cm/cm
Gal = cm/s**2
Mx = 10**-8*Wb
G = Mx/cm/cm
Oe = (10**3/(4*_math.pi))*A/m



"""Experimental and other useful units."""
from si.units.derived import *
from si.units.prefixed import *

# non-si units whose values in si units must be obtained experimentally, by si description (table 7)
# (xx) is the standard uncertainity of the last digits, units i would not recognize immediately are commented with name
eV = 1.60217653*10**-19*J # (14)
Da = 1.66053886*10**-25*kg # (28); Dalton
u = Da
ua = 1.49597870691*10**11*m # (6); astronomical unit
# natural units
c = 299792458*m/s # by definition
h__stroke = 1.05457168*10**-34*J*s # (18)
m_e = 9.1093826*10**-31*kg # (16); mass of an electron
# atomic units
e = 1.60217653*10**-19*C # (14)
a_0 = 0.5291772108*10**-10*m # (18); bohr radius
E_h = 4.35974417*10**-18*J # (75); hartree energy

# other
ly = 356.25*24*h*c # light year as recommended by the iau http://www.iau.org/Units.234.0.html



# encoding: utf8
"""Funny units."""
from si.units.derived import *
from si.units.prefixed import *
from si import prefixes as _prefixes

parsec=3.08567758128*10**16*m
attoparsec=_prefixes.atto*parsec

fortnight=h*24*14
microfortnight=_prefixes.micro*fortnight



from setuptools import setup

setup(
        name = "si",
        version = "0.1-2",
        packages = ["si", "si.units"],
        author = "chrysn",
        author_email = "chrysn@fsfe.org",
        description = "Module to represent SI units",
        license = "GPL",
        zip_safe = True,
        test_suite = "test.all",
        )
